# ngsolve-imports
from ngsolve import *
import ngs_petsc as petsc
from netgen.meshing import Mesh as NGMesh
import time
from math import pi

# initialize MPI
comm = mpi_world
rank = comm.rank

nx = 60
if rank==0:
    from ngsolve.meshes import MakeStructured3DMesh
    m1 = MakeStructured3DMesh(hexes=False, nx=nx, ny=nx, nz = nx) 
    ngm = m1.ngmesh
    if comm.size>1:
        ngm.Distribute(comm)
else:
    ngm = NGMesh.Receive(comm)
mesh = Mesh(ngm)

# reference solution
uex = sin(2*pi*x)*sin(2*pi*y)*sin(2*pi*z)
source = 12*pi**2*uex

# RT0-P0
V = HDiv(mesh, discontinuous=True)
Q = L2(mesh)
M = FacetFESpace(mesh, dirichlet=".*")
fes = FESpace([V, Q, M])
(u, p, phat), (v, q, qhat) = fes.TnT()

# RHS does not change either!
f = LinearForm (fes)
f += SymbolicLFI(source * q)
f.Assemble()

# neither does the BLF!
a = BilinearForm (fes, symmetric=False, condense=True)
n = specialcf.normal(mesh.dim)
a += (-u*v+p*div(v)+div(u)*q)*dx
a += (-u*n*qhat-v*n*phat)*dx(element_boundary=True)
a.Assemble()

### use PETSC KSP solver
opts = {"ksp_type":"cg", "ksp_atol":1e-30, "ksp_rtol":1e-8,
        "ksp_monitor" : "",
        "pc_type": "hypre", 
        "pc_hypre_type" : "boomeramg"}

t0 = time.time()    
mat_wrap = petsc.PETScMatrix(a.mat, freedofs=fes.FreeDofs(True), format=petsc.PETScMatrix.AIJ)
t1 = time.time()
ksp = petsc.KSP(mat=mat_wrap, name="someksp", petsc_options=opts, finalize=True)
t2 = time.time()
if rank==1:
  print("Rank %3d convertPetsc %.2e pre-setup: %.2e "%(rank,
    t1-t0, t2-t1))
#############

# solve the equation
gfu = GridFunction (fes)

f.vec.data += a.harmonic_extension_trans * f.vec
t0 = time.time()
gfu.vec.data = ksp*f.vec
if rank==1:
  print(gfu.vec.GetParallelStatus())
gfu.vec.Cumulate() # THIS LINE IS NECESSARY !!!
if rank==1:
  print(gfu.vec.GetParallelStatus())
t1 = time.time()
gfu.vec.data += a.harmonic_extension * gfu.vec
gfu.vec.data += a.inner_solve * f.vec


# error 
error = Integrate ( (gfu.components[1]-uex)**2 , mesh)

if rank==1:
  print("Rank %3d ndofs %4d Elapse: %.1e, error %.3e"%(rank,
    M.ndof, t1-t0, error))
